local T, scan_sector, check_number = ...

local handler = {
	["commands"] = function(data)
		print("\127FFFFFF"..T.help["commands"])
	end,
	
	["pos"] = function(data)
		if (T.hasmdi) then return end

		local x, y = data[1], data[2]
		local container = T.UI.display_container
		if (check_number(x) and check_number(y)) then
			if (x:match("^[+-]")) then x = container.cx + x end
			if (y:match("^[+-]")) then y = container.cy + y end
			T.x = x
			T.y = y
			T:SaveSettings()
			T:RebuildDisplay()

		else
			if (data[1] == "reset") then
				T.x = nil
				T.y = nil
				print("\127FFFFFFTargetlist position has been reset to default.")
				T:SaveSettings()
				T:RebuildDisplay()
			else
				local pos = container.cx.."x"..container.cy -- use the container positions since nil can't be concatenated
				if (not T.x) then pos = "unset (actual "..pos..")" end
				print("\127FFFFFFTargetlist position is "..pos)
			end
		end
	end,
	
	["fontscale"] = function(data)
		local newfontscale = check_number(data[1], T.defaults.fontscale.Lval, T.defaults.fontscale.Uval, T.defaults.fontscale.dp)
		if (newfontscale) then
			newfontscale = newfontscale < 0.1 and 0.1 or newfontscale
			T.fontscale = newfontscale
			print("\127FFFFFFFont scale has been changed to "..newfontscale..".")
			T:SaveSettings()
			T:RebuildDisplay()
		else
			if (data[1] == "reset") then
				T.fontscale = T.defaults.fontscale.val
				print("\127FFFFFFThe font scale has been been reset to "..T.fontscale..".")
				T:SaveSettings()
				T:RebuildDisplay()
			else
				print("\127FFFFFFThe current font scale is "..T.fontscale..".")
			end
		end
	end,
	
	["help"] = function(data)
		local word = table.remove(data, 1)
		if (word) then
			word = string.lower(tostring(word))
			if (type(T.help[word]) == "string") then
				print("\127FFFFFF"..T.help[word])
			end

		else
			print("\127FFFFFF"..T.help["help"])
		end
	end,
              
	["highlight"] = function(data)
		local r = check_number(data[1], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		local g = check_number(data[2], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		local b = check_number(data[3], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		local a = check_number(data[4], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		if (r and g and b) then
			a = a or 65
			T.current_target_color = string.format("%d %d %d %d *", r, g, b, a)
			print("\127FFFFFFThe highlight color has been changed to "..T.current_target_color:match("(.*) %*"))
			T:SaveSettings()
		else
			if (data[1] == "reset") then
				T.current_target_color = T.defaults.target_color.val
				print("\127FFFFFFThe highlight color has been reset to "..T.current_target_color:match("(.*) %*"))
				T:SaveSettings()
			else
				print("\127FFFFFFThe current highlight color is "..T.current_target_color:match("(.*) %*"))
			end
		end
	end,
	
	["rows"] = function(data)
		local newrows = check_number(data[1], T.defaults.list_count.Lval)
		if (newrows) then
			T.list_count = newrows
			print("\127FFFFFFThe number of rows has been changed to "..newrows..".")
			T:SaveSettings()
			T:RebuildDisplay()
		else
			if (data[1] == "reset") then
				T.list_count = T.defaults.list_count.val
				print("\127FFFFFFThe number of rows has been been reset to "..T.list_count)
				T:SaveSettings()
				T:RebuildDisplay()
			else
				print("\127FFFFFFThe current number of rows is "..T.list_count)
			end
		end
	end,

	["width"] = function(data)
		local newwidth = check_number(data[1], T.defaults.name_width.Lval)
		if (newwidth) then
			T.name_width = newwidth
			print("\127FFFFFFName width has been changed to "..newwidth)
			T:SaveSettings()
			T:RebuildDisplay()

		else
			if (data[1] == "reset") then
				T.name_width = nil
				print("\127FFFFFFName width has been reset.")
				T:SaveSettings()
				T:RebuildDisplay()
			else
				local width = T.display[1][2][3][2].w
				if (not T.name_width) then width = "unset (actual "..width..")" end
				print("\127FFFFFFThe current name width is "..width..".")
			end
		end
	end,

	["options"] = function()
		ShowDialog(T.optionsdlg)
	end,
}

local function cmd_processor(_, data)
	if (not data) then
		handler["options"]()
-- 		print("\127FFFFFF"..T.help["help"])
		return
	end
	local cmd = string.lower(table.remove(data, 1))
	if (type(handler[cmd]) == "function") then handler[cmd](data) end
end

local function exec_tab_func(func, ...)
	if (type(func) == "function") then
		local success, err = pcall(func, ...)
		if (not success) then
			console_print("ERROR: "..err)
			targetlist.API.UpdateDisplay(..., {{"ERROR", "ERROR", "ERROR"}})
		end
	end
end

function T:RefreshTab(is_targetlist_tab)
	ProcessEvent("TARGETLIST_TAB_CHANGED") -- Don't pass the identifier.  This could allow other plugins to interfere.
	if (not is_targetlist_tab) then
		local tab_tbl = T.tab_identifiers[T.tab_index]
		exec_tab_func(tab_tbl.ChangeFilter, tab_tbl, T.filter, T.faction_filter) -- execute after the event in case the event handlers need to do some intialization
	end
end


local function change_tab(i)
	local window = T.tab_window_size
	local offset = T.tab_window_offset
	local num_tabs = #T.tab_identifiers
	i = i < 0 and 0 or i > num_tabs and num_tabs or i
	local adj = i - T.tab_index

	if     (i == 1 and adj < i) then -- wrapping from right around to left
		offset = 0
	elseif (i == 0) then -- wrapping from left around to right
		i = num_tabs
		if (num_tabs > window) then
			offset = num_tabs - window
		end
	else
		-- Moving from right-to-left
		if (i < T.tab_index) then
			offset = i - 1
			if (offset > 0) then
				offset = offset - 1
			end

		-- Moving from left-to-right
		elseif (i > T.tab_index) then
			offset = i - window + 1
		end

		-- don't move the window off the left end
		if (offset >= num_tabs - window) then offset = num_tabs - window end
		-- don't move the window off the left end
		if (offset < 0) then offset = 0 end

	end

	T.tab_window_offset = offset
	T.tab_index = i

	-- Need to unregister the event before rebuilding, otherwise the wrong items could be selected
	if (T.tab_index > 1) then
		T.refresh_timer:Kill()
		UnregisterEvent(T.Events, "TARGET_CHANGED")
	elseif (T.tab_index == 1) then
		-- When switching back to the Targets tabe, the current target may have been changed.  So the current nodeID needs to be updated.
		T.current_target_nodeid = radar.GetRadarSelectionID()
		RegisterEvent(T.Events, "TARGET_CHANGED")
	end

	T.filter, T.faction_filter = T.tab_positions:GetFilterPositions(T.tab_index) -- legacy variable.  set to avoid rewriting core loops
	T:RebuildDisplay()

end


local function tab_next(_, data)
	local num = #T.tab_identifiers
	if (num == 1) then return end

	if (data) then
		data = tonumber(data[1]) % num
		if (data == 0) then data = num end
	end

	change_tab(data or ((T.tab_index % num) + 1))
end


local function tab_prev()
	local num = #T.tab_identifiers
	if (num == 1) then return end
	change_tab(((T.tab_index - 1) % num))
end

local function cycle_filter(_, data)
	local findex, fpage = T.tab_positions:GetFilterPositions(T.tab_index)
	if (not findex) then return end -- no filters to cycle

	local tab_tbl = T.tab_identifiers[T.tab_index]
	local filters = tab_tbl["filters"]

	T.UI.filters[findex]:SetActive(false)

	local r, num_filters = pcall(function() return #filters end)

	local max_filters = (r and num_filters or 1)
	if (data) then
		data = tonumber(data[1]) % max_filters
		if (data == 0) then data = max_filters end
	end

	findex = data or ((findex) % max_filters) + 1
	T.filter = T.tab_positions:SetFilterPositions(T.tab_index, findex, nil) -- legacy variable.  set to avoid rewriting core loops
	findex, fpage = T.tab_positions:GetFilterPositions(T.tab_index) -- findex changed,  fpage is now inaccurate.  get it again

	if (filters["link_filter_positions"]) then
		T.UI.filters[findex]:SetBGColor(filters[1]["pages"][fpage]["bgcolor"], tab_tbl["colorsource"])
	end
	T.UI.filters[findex]:SetActive(true)

	if (T.tab_index > 1) then
		exec_tab_func(tab_tbl.ChangeFilter, tab_tbl, findex, fpage)
	end

end

local function cycle_filter_page(_, p)
	local findex, fpage = T.tab_positions:GetFilterPositions(T.tab_index)
	local tab_tbl = T.tab_identifiers[T.tab_index]
	local filters = tab_tbl["filters"] --tab_tbl["filters"][findex]
	local filter = filters and filters[findex]
	local pages = filter and filter["pages"]
	local link = filters and filters["link_filter_positions"]

	if (not pages) then return end -- If there are no pages for this filter, then we're done

	local r, num_filter_pages = pcall(function() return #pages end)

	local max_pages = (r and num_filter_pages or 1)
	if (p) then
		p = tonumber(p[1]) % max_pages
		if (p == 0) then p = max_pages end
	end

	fpage = p or ((fpage) % (r and num_filter_pages or 1)) + 1
	T.faction_filter = T.tab_positions:SetFilterPositions(T.tab_index, nil, fpage) -- legacy variable.  set to avoid rewriting core loops

	local title
	local fgcolor
	local bgcolor
	title   = link and filter["title"] or pages[fpage]["title"]
	fgcolor = link and filters[1]["pages"][fpage]["fgcolor"] or pages[fpage]["fgcolor"] or filter["fgcolor"]
	bgcolor = link and filters[1]["pages"][fpage]["bgcolor"] or pages[fpage]["bgcolor"] or filter["bgcolor"]

	title = title or filter["title"]
	fgcolor = fgcolor
	bgcolor = bgcolor

	T.UI.filters[findex]:SetTitle(title)
	T.UI.filters[findex]:SetFGColor(fgcolor, tab_tbl.colorsource)
	T.UI.filters[findex]:SetBGColor(bgcolor, tab_tbl.colorsource)

	if (T.tab_index > 1) then
		exec_tab_func(tab_tbl.ChangeFilter, tab_tbl, findex, fpage)
	end
end

local function change_target(i, select_only, select_type)
	if (i == 0) then return end
	select_only = select_only or false

	if (T.tab_index == 1) then -- the Targets tab
		if (T.tab_target_list[i]) then
			if (select_only) then
				-- If selection is unlinked then a target change must be simulated to update the selection change
				ProcessEvent("TARGET_CHANGED", select_only, i)
			else
				radar.SetRadarSelection(T.tab_target_list[i]["nodeID"], T.tab_target_list[i]["objectID"])
			end
		end

	else -- all other tabs
		local tab_def = T.tab_identifiers[T.tab_index]

		local old_list_id
		if (select_only) then
			old_list_id = T.tab_positions:GetItemPosition(T.tab_index, "select")
		else
			old_list_id = T.tab_positions:GetItemPosition(T.tab_index, "active")
		end

		exec_tab_func(tab_def.ChangeTarget, tab_def, old_list_id, i, select_only, select_type or "select")
	end
end

local function is_unlinked(tab_status_only)
	local tab_link_status = T.tab_identifiers[T.tab_index]["controls"]
	local retval = tab_link_status ~= "linked" and (T.unlink_targeting or tab_link_status == "unlinked")

	if (tab_status_only) then
		retval = tab_link_status == "unlinked"
	end

	return retval
end

local function target_this()
	local listindex
	if (is_unlinked()) then
		listindex = T.tab_positions:GetItemPosition(T.tab_index, "select")
	else
		listindex = T.tab_positions:GetItemPosition(T.tab_index, "active")
	end
	change_target(listindex or 0, false)
end

local function find_target_in_range(i, list_end, step, wrap_to)
	local j = i + step

	if ((step < 0) and (j < list_end)) then
		j = wrap_to
	elseif (step > 0 and j > list_end) then
		j = wrap_to
	end

	while j ~= i do
		if (T.tab_target_list[j]["distance"] ~= 9999) then
			i = j
			break
		end

		j = j + step

		if ((step < 0) and (j < list_end)) then
			j = wrap_to
		elseif (step > 0 and j > list_end) then
			j = wrap_to
		end
	end

	return i
end

local function target_next()
	local num = #T.tab_target_list
	if (num == 0) then return end

	local is_unlinked = is_unlinked()

	local i = T.tab_positions:GetItemPosition(T.tab_index, "active")
	if (is_unlinked) then
		i = T.tab_positions:GetItemPosition(T.tab_index, "select") or i
	end
	i = i or 0

	i = (i + 1) % T.list_count

	if (i == 0) then i = T.list_count end
	if (i > num) then i = 1 end

	-- (N)PCs out of range have a distance of 9999.
	-- PCs can be pinned abvoe NPCs in range, which will cause a break in the display.
	-- Try to find another target in the list that is in range.
	-- Reset to the top of the list if none found.
	if (T.tab_index == 1 and T.tab_target_list[i]["distance"] == 9999) then
		i = find_target_in_range(i, num, 1, 1)
		if (i > T.list_count) then i = T.list_count end
	end

	change_target(i, is_unlinked, "scroll")
end

local function target_prev()
	local num = #T.tab_target_list
	if (num == 0) then return end

	local is_unlinked = is_unlinked()
	local i = T.tab_positions:GetItemPosition(T.tab_index, "active")
	if (is_unlinked) then
		i = T.tab_positions:GetItemPosition(T.tab_index, "select") or i
	end
	i = i or 0

	i = (i - 1) % T.list_count

	if (i == 0) then i = T.list_count end
	if (i > num) then i = num end

	-- (N)PCs out of range have a distance of 9999.
	-- PCs can be pinned abvoe NPCs in range, which will cause a break in the display.
	-- Try to find another target in the list that is in range.
	-- Reset to the bottom of the list if none found.
	if (T.tab_index == 1 and T.tab_target_list[i]["distance"] == 9999) then
		i = find_target_in_range(i, 1, -1, num)
		if (i > T.list_count) then i = T.list_count end
	end

	change_target(i, is_unlinked, "scroll")
end

local function pin_target(_, data)
	data = data and tonumber(data[1])

	if (T.tab_index == 1) then

		if (not T.tab_target_list[data]) then data = nil end

		local target_listindex = T.tab_positions:GetItemPosition(T.tab_index, "active")
		if (T.tab_index == 1 and data or target_listindex) then
			local charID = T.tab_target_list[data or target_listindex]["charID"]
			if (T.pinlist[charID]) then
				T.pinlist[charID] = nil
			else
				T.pinlist[charID] = os.time()
			end
		end
	else
		local tab_tbl = T.tab_identifiers[T.tab_index]
		exec_tab_func(tab_tbl.PinTarget, tab_tbl, data)
	end
end

local verb_color = {[true] = "\127FF0000", [false] = "\12700FF00"}

local verb = {
	{[true] = "only Hive", [false] = "all NPCs"},
	{[true] = "HIDING", [false] = "SHOWING"}, -- reversed logic because the variable is hide_turrets
	{[true] = "NPC ship types", [false] = "NPC names"},
	{[true] = "INVERTED", [false] = "NORMAL"},
	{[true] = "unlinked", [false] = "linked"},
}

local function toggle_hive_only()
	T.show_hive_only = not T.show_hive_only
	print(string.format("\127FFFFFFTargetList is now showing %s%s\127o on the NPC filter.", verb_color[T.show_hive_only], verb[1][T.show_hive_only]))
	T:SaveSettings()
end

local function toggle_friendly()
	T.hide_friendly = not T.hide_friendly
	print(string.format("\127FFFFFFTargetList is now %s%s\127o friendly targets.", verb_color[T.hide_friendly], verb[2][T.hide_friendly]))
	T:SaveSettings()
end

local function toggle_group()
	T.hide_group = not T.hide_group
	print(string.format("\127FFFFFFTargetList is now %s%s\127o group members.", verb_color[T.hide_group], verb[2][T.hide_group]))
	T:SaveSettings()
end

local function toggle_turrets()
	T.hide_turrets = not T.hide_turrets
	print(string.format("\127FFFFFFTargetList is now %s%s\127o turrets.", verb_color[T.hide_turrets], verb[2][T.hide_turrets]))
	T:SaveSettings()
end

local function toggle_npc_ship_types()
	T.show_npc_ship_types = not T.show_npc_ship_types
	print(string.format("\127FFFFFFTargetList is now showing %s%s\127o.", verb_color[not T.show_npc_ship_types], verb[3][T.show_npc_ship_types]))
	T:SaveSettings()
end

local function invert_faction_filter()
	T.invert_faction_filter = not T.invert_faction_filter
	print(string.format("\127FFFFFFTargetList faction filter is now %s%s\127o.", verb_color[T.invert_faction_filter], verb[4][T.invert_faction_filter]))
	iup.GetParent(T.UI.filters)[1].bgcolor = "255 0 0 "..(T.invert_faction_filter and 128 or 0)
	T:SaveSettings()
end

local function toggle_unlink_targeting()
	T.unlink_targeting = not T.unlink_targeting
	-- Clear variables related to the target selector  If switching to linked targeting they aren't needed.
	-- Got to make sure to clear the associated highlight and refresh the display, otherwise it could get reset to old values when the existing tab_target_list
	-- is used to re-update the display.  Since the most recent tab list is stored in the targetlist table, this will be cleared and updated separate from
	-- the tab operation.  This might cause a conflict of interest in how highlighting should be handled, but I'll cross the bridge when I come to it.
	local select_listindex = T.tab_positions:GetItemPosition(T.tab_index, "select")

	if (T.tab_index ~= 1 and select_listindex and not is_unlinked(true)) then
		-- Clear the relevant highlights for 3rd-party tabs.  Targets tab handles highlights more directly for speed
		T.tab_target_list[select_listindex]["highlight"] = nil
		T.tab_target_list[select_listindex]["type"] = nil
	end

	for tab_index in ipairs(T.tab_identifiers) do
		if (not is_unlinked(true)) then
			T.tab_positions:SetItemPosition(tab_index, "select", nil)
		end
	end
	T.current_select_nodeid = nil -- this is only relevant to the Targets tab
	T.UI.display_container.UpdateDisplay(T.tab_target_list, T.tab_index ~= 1)

	print(string.format("\127FFFFFFTargetList selection and targeting are now %s%s\127o.", verb_color[T.unlink_targeting], verb[5][T.unlink_targeting]))
	T:SaveSettings()

end

local function set_key_binds()
	for i = 1, T.list_count < 10 and T.list_count or 10 do
		gkinterface.GKProcessCommand("bind "..(i % 10).." target"..i)
	end
	gkinterface.GKProcessCommand("bind - cyclefilter")
	gkinterface.GKProcessCommand("bind = cyclefilterpage")
end

for i = 1, T.list_count < 10 and T.list_count or 10 do
	RegisterUserCommand("target"..i, change_target, i)
end

RegisterUserCommand("cyclefilter", cycle_filter)
RegisterUserCommand("cyclefilterpage", cycle_filter_page)
RegisterUserCommand("targetthis", target_this)
RegisterUserCommand("tabnext", tab_next)
RegisterUserCommand("tabprev", tab_prev)
RegisterUserCommand("invertfactionfilter", invert_faction_filter)
RegisterUserCommand("targetnext", target_next)
RegisterUserCommand("targetprev", target_prev)
RegisterUserCommand("pintarget", pin_target)
RegisterUserCommand("togglehiveonly", toggle_hive_only)
RegisterUserCommand("togglefriendly", toggle_friendly)
RegisterUserCommand("togglegroup", toggle_group)
RegisterUserCommand("toggleturrets", toggle_turrets)
RegisterUserCommand("togglenpcshiptypes", toggle_npc_ship_types)
RegisterUserCommand("togglelink", toggle_unlink_targeting)
RegisterUserCommand("settargetbinds", set_key_binds)
RegisterUserCommand("targetlist", cmd_processor)

return change_target
